@@ -280,7 +280,7 @@ HAVING = dsl.lift(TokenType.HAVING)
 HINT = dsl.lift(TokenType.HINT)
 IGNORE = dsl.lift(TokenType.IGNORE)
 ILIKE = dsl.lift(TokenType.ILIKE)
-ILIKE_ANY = dsl.lift(TokenType.ILIKE_ANY)
+
 IN = dsl.lift(TokenType.IN)
 INDEX = dsl.lift(TokenType.INDEX)
 INNER = dsl.lift(TokenType.INNER)
@@ -301,7 +301,7 @@ LANGUAGE = dsl.lift(TokenType.LANGUAGE)
 LATERAL = dsl.lift(TokenType.LATERAL)
 LEFT = dsl.lift(TokenType.LEFT)
 LIKE = dsl.lift(TokenType.LIKE)
-LIKE_ANY = dsl.lift(TokenType.LIKE_ANY)
+
 LIMIT = dsl.lift(TokenType.LIMIT)
 LIST = dsl.lift(TokenType.LIST)
 LOAD = dsl.lift(TokenType.LOAD)
